\ TDV is a package grouping a number of operations
\ for tridimensional  vectors
 
\ Copyright Paul KOPFF, CLAMART (FRANCE)
 
\ The tridimensional vectors are represented as triplets of 16 bits
\ cells, either on the stack, or within variable data structures.
 
\ Defining word for a 3D vector
 
: 3-VECTOR create 6 allot ;
 
\ Fetch and store functions
 
: 3V! ( adr --- x,y,z )
      6 0 DO DUP I + ROT SWAP ! 2 +LOOP DROP ;
 
: 3V@ ( x,y,z,adr --- )
      0 4 DO DUP I + @   SWAP  -2 +LOOP DROP ;
 
\ Stack manipulations
 
: 3VDUP   ( x,y,z --- x,y,z,x,y,z )
          2 PICK 2 PICK 2 PICK ;
 
: 3VOVER  ( x,y,z,n,n,n --- x,y,z,n,n,n,x,y,z )
          5 PICK 5 PICK 5 PICK ;
 
: 3VSWAP  ( x,y,z,n,n,n --- n,n,n,x,y,z )
          5 ROLL 5 ROLL 5 ROLL ;
 
: 3VROT   ( x,y,z,n,n,n,p,p,p --- n,n,n,p,p,p,x,y,z )
          8 ROLL 8 ROLL 8 ROLL ;
 
: 3VDROP  ( x,y,z --- )
          DROP DROP DROP ;
 
\ Addition and substraction
 
: 3V+     ( x,y,z,X,Y,Z --- x+X,y+Y,z+Z )
              3 ROLL + SWAP 3 ROLL + ROT 3 ROLL + SWAP ROT ;
 
: 3V-     ( x,y,z,X,Y,Z --- x-X,y-Y,z-Z )
       3VSWAP 3 ROLL - SWAP 3 ROLL - ROT 3 ROLL - SWAP ROT ;
 
\ The next two functions define the scalar and vector products
\ with their usual acceptance. However, in order to conserve the
\ scaling, we defined them like the */ and al. operators:
 
\ Their input paramreters are two vectors and a scalar scaling
\ factor. These data and the results are 16 bits numbers but the
\ intermediate computations are performed with 32 bits numbers.
 
\ Scalar product: the result is a scalar, ( V1 * V2 ) / s
 
: 3V*     ( x,y,z,X,Y,Z,s --- xX/s + yY/s + zZ/s )
  >R 3 ROLL *D 2SWAP 4 ROLL *D ROT 5 ROLL *D D+ D+ R> M/MOD SWAP DROP ;
 
\ Vector product: the result is a vector, ( V1 ^ V2 ) / s
 
: 3V^     ( x,y,z,X,Y,Z,s --- [yZ-Yz]/s,[zX-Zx]/s,[xY-Xy]/s )
  >R  4 PICK OVER *D 3 PICK 6 PICK *D D- R@ M/MOD SWAP DROP
      4 ROLL 4 PICK *D 3 ROLL 7 PICK *D D- R@ M/MOD SWAP DROP
      5 ROLL 3 ROLL *D 4 ROLL 5 ROLL *D D- R> M/MOD SWAP DROP ;
 
 
 
 
 
 
 
\ Square root and modulus of a vector
 
: SQRT ( d --- n,p | d = p*p + n )
      2DUP OR
      IF -32768
         BEGIN
            DUP >R DUP *D
            2OVER D- R@ UM/MOD
            SWAP DROP 2/
            R> SWAP ?DUP
         WHILE
            -
         REPEAT
         1- >R R@ R@ *D D-
         DROP R>
      THEN ;
 
: MODULUS ( x,y,z --- m )
      DUP *D ROT DUP *D D+ ROT DUP *D D+ SQRT SWAP DROP ;
 
 
 
\ Rotation: second order approximation !
\ x,y,z are the components of the vector to rotate
\ omegax,omegay,omegaz are the components of the polar vector
\ defining the axis and the angle of rotation
\ xr,yr,zr are the components of the resulting vector
 
: ROTATION ( x,y,z,omegax,omegay,omegaz --- xr,yr,zr )
      3VOVER 3VOVER 3VOVER MODULUS DUP >R
      3V^ 3VOVER R@ 3V^ 2/ ROT 2/ ROT 2/ ROT
      3VROT 3V+ 3VDUP 3VROT R> 3V^ 3V+ ;
 
 
 
\ The following function is a perspective projection
\ xp,yp,zp are the components of a point-of-view vector
\ i.e. the coordinates of the camera.
\ x,y,z are the components of the vector to transform
\ i.e. the coordinates of the current point.
\ xi,yi,zi are the components of the image vector
\ i.e. the coordinates of the intersection of the
\ line going from the camera to the current point, with
\ the plane containing the origin of coordinates and
\ orthogonal to the point-of-view vector.
 
\ Hence the formula of the transform:
 
\            ->   ->     ->   ->       ->   ->
\            OI = OP ^ ( OP ^ MP ) / ( OP * MP )
 
: PERSPECTIVE ( xp,yp,zp,x,y,z --- xi,yi,zi )
      3VOVER 3VDUP 3VROT 3V-
      3VOVER 3VOVER 3VOVER MODULUS DUP >R
      3V* 3V^ R> 3V^ ;
 
 